from pydantic import BaseModel
from rich import get_console
from turms.processor.base import Processor


class BlackProcessorConfig(BaseModel):
    pass


class BlackProcessor(Processor):
    def __init__(self, config=None, **data):

        self.plugin_config = config or BlackProcessorConfig(**data)

    def run(self, gen_file: str):
        try:
            from black import format_str, FileMode

            return format_str(gen_file, mode=FileMode())
        except Exception as e:
            get_console().log(
                "Wasn't able to use the Blackprocessor. Please Make sure to install black."
            )

        return gen_file



from graphql import (
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLScalarType,
    GraphQLWrappingType,
)
from turms.globals import INPUTTYPE_CLASS_MAP
from turms.plugins.base import Plugin
import ast
from typing import List, Optional
from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from turms.plugins.base import Plugin
from pydantic import BaseModel
from graphql.type.definition import (
    GraphQLEnumType,
)

from turms.utils import get_scalar_equivalent


class InputsPluginConfig(BaseModel):
    inputtype_bases: List[str] = ["turms.types.object.GraphQLInputObject"]
    skip_underscore: bool = True
    prepend: str = ""
    append: str = "Input"


def generate_input_annotation(
    type: GraphQLInputType,
    config: GeneratorConfig,
    plugin_config: InputsPluginConfig,
    is_optional=True,
):
    if isinstance(type, GraphQLScalarType):
        if is_optional:
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Name(
                    id=get_scalar_equivalent(type.name, config), ctx=ast.Load()
                ),
            )
        else:
            return ast.Name(
                id=get_scalar_equivalent(type.name, config),
                ctx=ast.Load(),
            )

    if isinstance(type, GraphQLInputObjectType):
        if is_optional:
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Constant(
                    value=f"{config.prepend_input}{type.name}{config.append_input}",
                ),
            )
        return ast.Constant(
            value=f"{config.prepend_input}{type.name}{config.append_input}",
        )

    if isinstance(type, GraphQLEnumType):
        if is_optional:
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Constant(
                    value=f"{config.prepend_enum}{type.name}{config.append_enum}",
                ),
            )
        return ast.Constant(
            value=f"{config.prepend_enum}{type.name}{config.append_enum}",
        )
    if isinstance(type, GraphQLNonNull):
        return generate_input_annotation(
            type.of_type, config, plugin_config, is_optional=False
        )

    if isinstance(type, GraphQLList):
        if is_optional:
            return ast.Subscript(
                value=ast.Name("Optional", ctx=ast.Load()),
                slice=ast.Subscript(
                    value=ast.Name("List", ctx=ast.Load()),
                    slice=generate_input_annotation(
                        type.of_type, config, plugin_config, is_optional=True
                    ),
                    ctx=ast.Load(),
                ),
            )

        return ast.Subscript(
            value=ast.Name("List", ctx=ast.Load()),
            slice=generate_input_annotation(
                type.of_type, config, plugin_config, is_optional=True
            ),
            ctx=ast.Load(),
        )

    raise NotImplementedError(f"Unknown input type {type}")


def generate_inputs(
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: InputsPluginConfig,
):

    tree = []

    inputobjects_type = {
        key: value
        for key, value in client_schema.type_map.items()
        if isinstance(value, GraphQLInputObjectType)
    }

    self_referential = []

    for key, type in inputobjects_type.items():

        if plugin_config.skip_underscore and key.startswith("_"):
            continue

        name = f"{config.prepend_input}{key}{config.append_input}"
        fields = [ast.Expr(value=ast.Constant(value=type.description))]

        for value_key, value in type.fields.items():

            if isinstance(value.type, GraphQLInputObjectType):
                self_referential.append(name)

            assign = ast.AnnAssign(
                target=ast.Name(value_key, ctx=ast.Store()),
                annotation=generate_input_annotation(
                    value.type, config, plugin_config, is_optional=True
                ),
                simple=1,
            )

            potential_comment = (
                value.description
                if not value.deprecation_reason
                else f"DEPRECATED: {value.description}"
            )

            if potential_comment:
                fields += [
                    assign,
                    ast.Expr(value=ast.Constant(value=potential_comment)),
                ]

            else:
                fields += [assign]

        INPUTTYPE_CLASS_MAP[key] = name
        tree.append(
            ast.ClassDef(
                name,
                bases=[
                    ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                    for base in plugin_config.inputtype_bases
                ],
                decorator_list=[],
                keywords=[],
                body=fields,
            )
        )

    for input_name in self_referential:
        tree.append(
            ast.Expr(
                value=ast.Call(
                    func=ast.Attribute(
                        value=ast.Name(
                            id=input_name,
                            ctx=ast.Load(),
                        ),
                        attr="update_forward_refs",
                        ctx=ast.Load(),
                    ),
                    keywords=[],
                    args=[],
                )
            )
        )

    return tree


class InputsPlugin(Plugin):
    def __init__(self, config=None, **data):
        self.plugin_config = config or InputsPluginConfig(**data)

    def generate_imports(
        self, config: GeneratorConfig, client_schema: GraphQLSchema
    ) -> List[ast.AST]:
        imports = []

        all_bases = self.plugin_config.inputtype_bases

        for item in all_bases:
            imports.append(
                ast.ImportFrom(
                    module=".".join(item.split(".")[:-1]),
                    names=[ast.alias(name=item.split(".")[-1])],
                    level=0,
                )
            )

        return imports

    def generate_body(
        self, client_schema: GraphQLSchema, config: GeneratorConfig
    ) -> List[ast.AST]:

        return generate_inputs(client_schema, config, self.plugin_config)



from abc import abstractmethod
import ast
from typing import Callable, List, Optional
from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from turms.globals import FRAGMENT_CLASS_MAP, FRAGMENT_DOCUMENT_MAP
from turms.parser.recurse import recurse_annotation, type_field_node
from turms.plugins.base import Plugin
from pydantic import BaseModel
from graphql.language.ast import FragmentDefinitionNode
from turms.utils import (
    NoDocumentsFoundError,
    generate_typename_field,
    get_additional_bases_for_type,
    parse_documents,
)
import re
from graphql import (
    FieldNode,
    FragmentSpreadNode,
    GraphQLInterfaceType,
    InlineFragmentNode,
    language,
    parse,
    get_introspection_query,
    validate,
)


from graphql.utilities.type_info import TypeInfo, get_field_def
import logging


logger = logging.getLogger(__name__)


class FragmentsPluginConfig(BaseModel):
    fragment_bases: List[str] = ["turms.types.object.GraphQLObject"]
    fragments_glob: Optional[str]


def generate_fragment(
    f: FragmentDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: FragmentsPluginConfig,
):
    tree = []
    fields = []
    type = client_schema.get_type(f.type_condition.name.value)

    name = (
        f"{config.prepend_fragment}{f.name.value.capitalize()}{config.append_fragment}"
    )

    if isinstance(type, GraphQLInterfaceType):
        mother_class_fields = []
        base_fragment_name = f"{name}Base"

        if type.description:
            mother_class_fields.append(
                ast.Expr(value=ast.Constant(value=type.description))
            )

        mother_class_fields += [
            ast.AnnAssign(
                target=ast.Name(id="typename", ctx=ast.Store()),
                annotation=ast.Subscript(
                    value=ast.Name(id="Optional", ctx=ast.Load()),
                    slice=ast.Name("str", ctx=ast.Load()),
                ),
                value=ast.Call(
                    func=ast.Name(id="Field", ctx=ast.Load()),
                    args=[],
                    keywords=[
                        ast.keyword(arg="alias", value=ast.Constant(value="__typename"))
                    ],
                ),
                simple=1,
            )
        ]

        for sub_node in f.selection_set.selections:

            if isinstance(sub_node, FieldNode):

                if sub_node.name.value == "__typename":
                    continue

                field_type = type.fields[sub_node.name.value]
                mother_class_fields += type_field_node(
                    sub_node,
                    field_type,
                    client_schema,
                    config,
                    tree,
                    parent_name=base_fragment_name,
                )

        additional_bases = get_additional_bases_for_type(type.name, config)

        mother_class = ast.ClassDef(
            base_fragment_name,
            bases=[
                ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                for base in config.interface_bases
            ]
            + additional_bases,  # Todo: fill with base
            decorator_list=[],
            keywords=[],
            body=mother_class_fields if mother_class_fields else [ast.Pass()],
        )

        tree.append(mother_class)

        union_class_names = []

        for sub_node in f.selection_set.selections:

            if isinstance(sub_node, FragmentSpreadNode):

                spreaded_name = f"{base_fragment_name}{sub_node.name.value}"

                cls = ast.ClassDef(
                    spreaded_name,
                    bases=[
                        ast.Name(
                            id=FRAGMENT_CLASS_MAP[sub_node.name.value], ctx=ast.Load()
                        ),
                        ast.Name(id=base_fragment_name, ctx=ast.Load()),
                    ],
                    decorator_list=[],
                    keywords=[],
                    body=[ast.Pass()],
                )

                tree.append(cls)
                union_class_names.append(spreaded_name)

            if isinstance(sub_node, InlineFragmentNode):
                inline_name = (
                    f"{base_fragment_name}{sub_node.type_condition.name.value}Fragment"
                )
                inline_fragment_fields = []

                inline_fragment_fields += [
                    generate_typename_field(sub_node.type_condition.name.value)
                ]

                for sub_sub_node in sub_node.selection_set.selections:

                    if isinstance(sub_sub_node, FieldNode):
                        sub_sub_node_type = client_schema.get_type(
                            sub_node.type_condition.name.value
                        )

                        if sub_sub_node.name.value == "__typename":
                            continue

                        field_type = sub_sub_node_type.fields[sub_sub_node.name.value]
                        inline_fragment_fields += type_field_node(
                            sub_sub_node,
                            field_type,
                            client_schema,
                            config,
                            tree,
                            parent_name=inline_name,
                        )

                additional_bases = get_additional_bases_for_type(
                    sub_node.type_condition.name.value, config
                )
                cls = ast.ClassDef(
                    inline_name,
                    bases=[
                        ast.Name(id=base_fragment_name, ctx=ast.Load()),
                    ],
                    decorator_list=[],
                    keywords=[],
                    body=inline_fragment_fields,
                )

                tree.append(cls)
                union_class_names.append(inline_name)

        union_class_names.append(base_fragment_name)

        FRAGMENT_DOCUMENT_MAP[f.name.value] = language.print_ast(f)
        FRAGMENT_CLASS_MAP[f.name.value] = name

        if len(union_class_names) > 1:
            slice = ast.Tuple(
                elts=[
                    ast.Name(id=clsname, ctx=ast.Load())
                    for clsname in union_class_names
                ],
                ctx=ast.Load(),
            )
            tree.append(
                ast.Assign(
                    targets=[ast.Name(id=name, ctx=ast.Store())],
                    value=ast.Subscript(
                        value=ast.Name("Union", ctx=ast.Load()),
                        slice=slice,
                    ),
                )
            )

        return tree

    name = f"{config.prepend_fragment}{f.name.value}{config.append_fragment}"

    additional_bases = get_additional_bases_for_type(
        f.type_condition.name.value, config
    )

    fields += [generate_typename_field(type.name)]

    for field in f.selection_set.selections:

        if isinstance(field, FragmentDefinitionNode):
            continue

        if isinstance(field, FragmentSpreadNode):
            continue

        field_definition = get_field_def(client_schema, type, field)
        assert field_definition, "Couldn't find field definition"

        target = (
            field.alias.value
            if hasattr(field, "alias") and field.alias
            else field.name.value
        )

        fields += type_field_node(
            field,
            field_definition,
            client_schema,
            config,
            tree,
            parent_name=name,
        )

    FRAGMENT_DOCUMENT_MAP[f.name.value] = language.print_ast(f)
    FRAGMENT_CLASS_MAP[f.name.value] = name
    tree.append(
        ast.ClassDef(
            name,
            bases=additional_bases
            + [
                ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                for base in plugin_config.fragment_bases
            ],
            decorator_list=[],
            keywords=[],
            body=fields,
        )
    )
    return tree


class FragmentsPlugin(Plugin):
    def __init__(self, config=None, **data):
        self.plugin_config = config or FragmentsPluginConfig(**data)

    def generate_imports(
        self, config: GeneratorConfig, client_schema: GraphQLSchema
    ) -> List[ast.AST]:
        imports = []

        all_bases = {base for base in self.plugin_config.fragment_bases}

        for item in all_bases:
            imports.append(
                ast.ImportFrom(
                    module=".".join(item.split(".")[:-1]),
                    names=[ast.alias(name=item.split(".")[-1])],
                    level=0,
                )
            )

        return imports

    def generate_body(
        self, client_schema: GraphQLSchema, config: GeneratorConfig
    ) -> List[ast.AST]:

        plugin_tree = []

        try:
            documents = parse_documents(
                client_schema, self.plugin_config.fragments_glob or config.documents
            )
        except NoDocumentsFoundError as e:
            logger.exception(e)
            return plugin_tree

        definitions = documents.definitions
        fragments = [
            node for node in definitions if isinstance(node, FragmentDefinitionNode)
        ]

        for fragment in fragments:
            plugin_tree += generate_fragment(
                fragment, client_schema, config, self.plugin_config
            )

        return plugin_tree
